local _G = _G

local ADDON = ...

local Addon = _G.BrokerXPBar

-- rearrangement of blizzard's frames
local Jostle	= LibStub:GetLibrary("LibJostle-3.0", true)

local LibSharedMedia = LibStub("LibSharedMedia-3.0", true)

-- constants
local SPARK_LEN_MIN  = 28
local SPARK_LEN_MAX  = 128

local SPARK2_LEN_MIN = 8
local SPARK2_LEN_MAX = 32

local TEX_BAR    = "Interface\\AddOns\\" .. ADDON .. "\\Textures\\texture.tga"
local TEX_BORDER = "Interface\\AddOns\\" .. ADDON .. "\\Textures\\border.tga"
local TEX_SPARK  = "Interface\\AddOns\\" .. ADDON .. "\\Textures\\glow.tga"
local TEX_SPARK2 = "Interface\\AddOns\\" .. ADDON .. "\\Textures\\glow2.tga"

-- the bar frame
local Bar = {
	-- layout
	anchored    = false,
	anchorFrame = nil,
	horizontal  = true,
	
	-- data
	progress = {
		XP          = 0,
		Rested      = 0,
		Reputation  = 0
	},
	
	-- settings
	settings = {
		Frame     = "dummy", 
		Inside    = false, 
		Inverse   = false, 
		Location  = "Top", 
		Jostle    = false, 
		Shadow    = true, 
		ShowXP    = true, 
		ShowRep   = true, 
		Spark     = 1, 
		Strata    = "HIGH", 
		Thickness = 2, 
		xOffset   = 0, 
		yOffset   = 0,
		Texture   = TEX_BAR,
	},
	
	-- colors
	colors = {
		XP        = {r=0.0, g=0.4, b=0.9, a=1},
		Rest      = {r=1.0, g=0.2, b=1.0, a=1},
		None      = {r=0.0, g=0.0, b=0.0, a=1},
		Rep       = {r=1.0, g=0.2, b=1.0, a=1},
		NoRep     = {r=0.0, g=0.0, b=0.0, a=1},
	},
}

Addon.Bar = Bar

-- utilities
local function CalcSparkLength(min, max, length)
	if length < max then
		return min + (length*(max-min)/max)
	else
		return max
	end
end

function Bar:Initialize()
	local r, g, b, a
	
	local thickness = self:GetSetting("Thickness")
	local intensity = self:GetSetting("Spark")
	local strata    = self:GetSetting("Strata")
	
	-- bar setup
	local MainFrame = CreateFrame("Frame", ADDON.."_Bar_MainFrame", UIParent)
	MainFrame:SetFrameStrata(strata)
	MainFrame:SetHeight(2 * thickness + 5)	
	MainFrame:SetWidth(2 * thickness + 5)	
	
	local XPBar = CreateFrame("Frame", ADDON.."_Bar_XPBar", MainFrame)
	XPBar:SetFrameStrata(strata)
	XPBar:SetHeight(thickness)

	r, g, b, a = self:GetColor("XP")
	
	local tex = XPBar:CreateTexture(ADDON.."_Tex_BarTex", "ARTWORK")
	tex:SetTexture(TEX_BAR)
	tex:SetVertexColor(r, g, b, a)
	tex:ClearAllPoints()
	tex:Show()
	
	local spark = XPBar:CreateTexture(ADDON.."_Tex_XPSpark", "OVERLAY")
	spark:SetTexture(TEX_SPARK)
	spark:SetAlpha(intensity)
	spark:SetWidth(128)
	spark:SetHeight(thickness * 2 + 12)
	spark:SetVertexColor(r, g, b, intensity)
	spark:SetBlendMode("ADD")

	local spark2 = XPBar:CreateTexture(ADDON.."_Tex_XPSpark2", "OVERLAY")
	spark2:SetTexture(TEX_SPARK2)
	spark2:SetAlpha(intensity)
	spark2:SetWidth(32)
	spark2:SetHeight(thickness * 8)
	spark2:SetBlendMode("ADD")

	r, g, b, a = self:GetColor("Rest")
	
	local restex = XPBar:CreateTexture(ADDON.."_Tex_RestedXPTex", "BORDER")
	restex:SetTexture(TEX_BAR)
	restex:SetVertexColor(r, g, b, a)
	restex:ClearAllPoints()
	restex:Show()

	r, g, b, a = self:GetColor("None")
	
	local notex = XPBar:CreateTexture(ADDON.."_Tex_NoXPTex", "BACKGROUND")
	notex:SetTexture(TEX_BAR)
	notex:SetVertexColor(r, g, b, a)
	notex:ClearAllPoints()
	notex:Show()
	notex:SetAllPoints(XPBar)
	
	local Rep = CreateFrame("Frame", ADDON.."_Bar_RepBar", MainFrame)
	Rep:SetFrameStrata(strata)
	Rep:SetHeight(thickness)

	r, g, b, a = self:GetColor("Rep")
	
	local reptex = Rep:CreateTexture(ADDON.."_Tex_RepTex", "ARTWORK")
	reptex:SetTexture(TEX_BAR)
	reptex:SetVertexColor(r, g, b, a)
	reptex:ClearAllPoints()
	reptex:Show()

	local rspark = Rep:CreateTexture(ADDON.."_Tex_RepSpark", "OVERLAY")
	rspark:SetTexture(TEX_SPARK)
	rspark:SetAlpha(intensity)
	rspark:SetWidth(128)
	rspark:SetHeight(thickness * 2 + 12)
	rspark:SetVertexColor(r, g, b, intensity)
	rspark:SetBlendMode("ADD")

	local rspark2 = Rep:CreateTexture(ADDON.."_Tex_RepSpark2", "OVERLAY")
	rspark2:SetTexture(TEX_SPARK2)
	rspark2:SetAlpha(intensity)
	rspark2:SetWidth(32)
	rspark2:SetHeight(thickness * 8)
	rspark2:SetBlendMode("ADD")

	r, g, b, a = self:GetColor("NoRep")
	
	local noreptex = Rep:CreateTexture(ADDON.."_Tex_NoRepTex", "BACKGROUND")
	noreptex:SetTexture(TEX_BAR)
	noreptex:SetVertexColor(r, g, b, a)
	noreptex:ClearAllPoints()
	noreptex:SetAllPoints(Rep)
	noreptex:Show()

	local Border = CreateFrame("Frame", ADDON.."_Bar_Border", MainFrame)
	Border:SetFrameStrata(strata)
	Border:SetHeight(5)
	
	local bordtex = Border:CreateTexture(ADDON.."_Tex_BorderTex", "BACKGROUND")
	bordtex:SetTexture(TEX_BORDER)
	bordtex:SetVertexColor(0, 0, 0, 1)
	bordtex:ClearAllPoints()
	bordtex:SetAllPoints(Border)

	self.MainFrame   = MainFrame
	self.XPBar       = XPBar
	self.XPBarTex    = tex
	self.Spark       = spark
	self.Spark2      = spark2
	self.NoXPTex     = notex
	self.RestedXPTex = restex
	self.RepBar      = Rep
	self.RepBarTex   = reptex
	self.RepSpark    = rspark
	self.RepSpark2   = rspark2
	self.NoRepTex    = noreptex
	self.Border      = Border
	self.BorderTex   = bordtex	
end

function Bar:Reanchor()
	local point, relpoint, y
	local sy = 0
	
	self.anchorFrame = self:GetSetting("Frame") and getglobal(self:GetSetting("Frame")) or nil
	
	if self.anchorFrame == nil then
		self:Hide()
		return 
	end
	
	local location = self:GetSetting("Location")
	local inside   = self:GetSetting("Inside")
	local inverse  = self:GetSetting("Inverse")
		
	if location == "Top" or location == "Bottom" then
		self.horizontal = true
	else
		self.horizontal = false
	end

	-- assign anchor points
	if location == "Top" then
		relpoint = "TOPLEFT"
	elseif location == "Right" then
		relpoint = "BOTTOMRIGHT"
	else
		relpoint = "BOTTOMLEFT"
	end

	if self.horizontal then
		if (location == "Top" and (not inside) ) or
			(location ~= "Top" and inside) then
			point = "BOTTOMLEFT"
			self.BorderTex:SetTexCoord(1, 0, 1, 0)
			y = 1
		else
			point = "TOPLEFT"
			self.BorderTex:SetTexCoord(1, 0, 0, 1)
			y = -1
		end

		self.XPBarTex:SetTexCoord(1, 0, 1, 1, 0, 0, 0, 1)
		self.NoXPTex:SetTexCoord(1, 0, 1, 1, 0, 0, 0, 1)
		self.RestedXPTex:SetTexCoord(1, 0, 1, 1, 0, 0, 0, 1)
		self.Spark:SetTexCoord(0, 1, 0, 0, 1, 1, 1, 0)
		self.Spark2:SetTexCoord(0, 1, 0, 0, 1, 1, 1, 0)

		self.RepBarTex:SetTexCoord(1, 0, 1, 1, 0, 0, 0, 1)
		self.NoRepTex:SetTexCoord(1, 0, 1, 1, 0, 0, 0, 1)
		self.RepSpark:SetTexCoord(0, 1, 0, 0, 1, 1, 1, 0)
		self.RepSpark2:SetTexCoord(0, 1, 0, 0, 1, 1, 1, 0)
	else
		if (location == "Right" and (not inside) ) or
			(location ~= "Right" and inside) then
			point = "BOTTOMLEFT"
			self.BorderTex:SetTexCoord(0, 0, 1, 0, 0, 1, 1, 1)
			y = 1
		else
			point = "BOTTOMRIGHT"
			self.BorderTex:SetTexCoord(1, 1, 0, 1, 1, 0, 0, 0)
			y = -1
		end

		self.XPBarTex:SetTexCoord(0, 0, 1, 0, 0, 1, 1, 1)
		self.NoXPTex:SetTexCoord(0, 0, 1, 0, 0, 1, 1, 1)
		self.RestedXPTex:SetTexCoord(0, 0, 1, 0, 0, 1, 1, 1)
		self.Spark:SetTexCoord(1, 1, 0, 1, 1, 0, 0, 0)
		self.Spark2:SetTexCoord(1, 1, 0, 1, 1, 0, 0, 0)

		self.RepBarTex:SetTexCoord(0, 0, 1, 0, 0, 1, 1, 1)
		self.NoRepTex:SetTexCoord(0, 0, 1, 0, 0, 1, 1, 1)
		self.RepSpark:SetTexCoord(1, 1, 0, 1, 1, 0, 0, 0)
		self.RepSpark2:SetTexCoord(1, 1, 0, 1, 1, 0, 0, 0)
	end
	
	-- detach from old frame
	self.MainFrame:ClearAllPoints()
	self.XPBar:ClearAllPoints()
	self.RepBar:ClearAllPoints()
	self.Border:ClearAllPoints()
	
	local thickness = self:GetSetting("Thickness")
	
	-- attach to new frame
	local xOffset = self:GetSetting("xOffset")
	local yOffset = self:GetSetting("yOffset")
	
	self:AttachBarToFrame(self.MainFrame, point, self.anchorFrame, relpoint, xOffset, yOffset)
	
	if self:GetSetting("ShowXP") then
		local offset = 0
		
		if inverse and self:GetSetting("ShowRep") then
			offset = y * thickness
		end
		
		if self.horizontal then
			xOffset = 0
			yOffset = offset
		else
			xOffset = offset
			yOffset = 0
		end
		
		self:AttachBarToFrame(self.XPBar, point, self.MainFrame, point, xOffset, yOffset)
		
		sy = sy + y
	end
	
	if self:GetSetting("ShowRep") then
		local offset = 0
		
		if not inverse and self:GetSetting("ShowXP") then
			offset = y * thickness
		end
		
		if self.horizontal then
			xOffset = 0
			yOffset = offset
		else
			xOffset = offset
			yOffset = 0
		end

		self:AttachBarToFrame(self.RepBar, point, self.MainFrame, point, xOffset, yOffset)

		sy = sy + y
	end
		
	if self:GetSetting("Shadow") then
		if self.horizontal then
			xOffset = 0
			yOffset = sy * thickness
		else
			xOffset = sy * thickness
			yOffset = 0
		end

		self:AttachBarToFrame(self.Border, point, self.MainFrame, point, xOffset, yOffset)
	end
	
	-- reset tex points
	self.Spark:ClearAllPoints()
	self.Spark2:ClearAllPoints()

	self.XPBarTex:ClearAllPoints()
	self.XPBarTex:SetPoint("TOP",    self.XPBar, "TOP")
	self.XPBarTex:SetPoint("BOTTOM", self.XPBar, "BOTTOM")
	self.XPBarTex:SetPoint("LEFT",   self.XPBar, "LEFT")
	self.XPBarTex:SetPoint("RIGHT",  self.XPBar, "RIGHT")
	
	self.RestedXPTex:ClearAllPoints()
	self.RestedXPTex:SetPoint("TOP",    self.XPBar, "TOP")
	self.RestedXPTex:SetPoint("BOTTOM", self.XPBar, "BOTTOM")
	self.RestedXPTex:SetPoint("LEFT",   self.XPBar, "LEFT")
	self.RestedXPTex:SetPoint("RIGHT",  self.XPBar, "RIGHT")
	
	self.RepSpark:ClearAllPoints()
	self.RepSpark2:ClearAllPoints()

	self.RepBarTex:ClearAllPoints()
	self.RepBarTex:SetPoint("TOP",    self.RepBar, "TOP")
	self.RepBarTex:SetPoint("BOTTOM", self.RepBar, "BOTTOM")
	self.RepBarTex:SetPoint("LEFT",   self.RepBar, "LEFT")
	self.RepBarTex:SetPoint("RIGHT",  self.RepBar, "RIGHT")
	
	-- restore minimal bar size	
	if self.horizontal then
		self.XPBar:SetWidth(thickness)

		self.Spark:SetWidth(128)
		self.Spark2:SetWidth(32)

		self.RepBar:SetWidth(thickness)
	
		self.RepSpark:SetWidth(128)
		self.RepSpark2:SetWidth(32)
	else
		self.XPBar:SetHeight(thickness)
	
		self.Spark:SetHeight(128)
		self.Spark2:SetHeight(32)

		self.RepBar:SetHeight(thickness)
	
		self.RepSpark:SetHeight(128)
		self.RepSpark2:SetHeight(32)
	end
	
	self.Border:SetWidth(8)
	self.Border:SetHeight(8)
	
	self:SetThickness(thickness)

	-- refresh settings
	self:UpdateStrata()
	self:UpdateSparkIntensity()
	self:UpdateTexture()
	
	-- (re)anchoring finished
	self.anchored = true	
	
	self:Show()
	
	self:Update()

	if Jostle then Jostle:Refresh() end
end

function Bar:AttachBarToFrame(bar, point, frame, relpoint, xOffset, yOffset)
	bar:SetParent(frame)
	bar:SetPoint(point, frame, relpoint, xOffset, yOffset)
end

function Bar:TextureSetPoint(tex, bar, point, offset)
	local xOffset, yOffset
	
	if self.horizontal then
		xOffset = offset
		yOffset = 0
	else
		xOffset = 0
		yOffset = offset
	end
	
	tex:SetPoint(point, bar, point, xOffset, yOffset)
end

function Bar:Update()
	local length
	local front, back, barlength

	if self.anchorFrame == nil then
		return
	end
	
	if self.horizontal then
		front = "LEFT"
		back  = "RIGHT"
	
		barlength = self.anchorFrame:GetWidth()

		-- adjust to possible changes in parent frame dimensions
		self.MainFrame:SetWidth(barlength)
		self.XPBar:SetWidth(barlength)
		self.RepBar:SetWidth(barlength)
		self.Border:SetWidth(barlength)
	else
		front = "BOTTOM"
		back  = "TOP"
	
		barlength = self.anchorFrame:GetHeight()

		-- adjust to possible changes in parent frame dimensions
		self.MainFrame:SetHeight(barlength)
		self.XPBar:SetHeight(barlength)
		self.RepBar:SetHeight(barlength)
		self.Border:SetHeight(barlength)
	end

	if self:GetSetting("ShowXP") then
		length = self.progress.XP * barlength
		
		self:TextureSetPoint(self.XPBarTex, self.XPBar, back, length-barlength)
		self:TextureSetPoint(self.RestedXPTex, self.XPBar, front, length)

		self:TextureSetPoint(self.Spark, self.XPBar, back, length-barlength+3)
		self:TextureSetPoint(self.Spark2, self.XPBar, back, length-barlength+3)
		
		-- resize spark to avoid excessive overlapping
		if self.horizontal then
			self.Spark:SetWidth(CalcSparkLength(SPARK_LEN_MIN, SPARK_LEN_MAX, length))
			self.Spark2:SetWidth(CalcSparkLength(SPARK2_LEN_MIN, SPARK2_LEN_MAX, length))	
		else
			self.Spark:SetHeight(CalcSparkLength(SPARK_LEN_MIN, SPARK_LEN_MAX, length))
			self.Spark2:SetHeight(CalcSparkLength(SPARK2_LEN_MIN, SPARK2_LEN_MAX, length))	
		end
		
		if self.progress.XP + self.progress.Rested > 1 then
			length = barlength
		else
			length = length + self.progress.Rested * barlength
		end

		self:TextureSetPoint(self.RestedXPTex, self.XPBar, back, length-barlength)
	end

	if self:GetSetting("ShowRep") then
		length = self.progress.Reputation * barlength
		
		self:TextureSetPoint(self.RepBarTex, self.RepBar, back, length-barlength)

		self:TextureSetPoint(self.RepSpark, self.RepBar, back, length-barlength+3)
		self:TextureSetPoint(self.RepSpark2, self.RepBar, back, length-barlength+3)
		
		-- resize spark to avoid excessive overlapping
		if self.horizontal then
			self.RepSpark:SetWidth(CalcSparkLength(SPARK_LEN_MIN, SPARK_LEN_MAX, length))
			self.RepSpark2:SetWidth(CalcSparkLength(SPARK2_LEN_MIN, SPARK2_LEN_MAX, length))	
		else
			self.RepSpark:SetHeight(CalcSparkLength(SPARK_LEN_MIN, SPARK_LEN_MAX, length))
			self.RepSpark2:SetHeight(CalcSparkLength(SPARK2_LEN_MIN, SPARK2_LEN_MAX, length))	
		end
	end
end

function Bar:Show()
	local height = 0
	
	self:Hide()

	local thickness = self:GetSetting("Thickness")
	local jostle    = self:GetSetting("Jostle")
	local location  = self:GetSetting("Location")
	
	if self:GetSetting("ShowXP") then
		self.XPBar:Show()
		self.Spark:Show()
		self.Spark2:Show()
		
		height = height + thickness

		-- register for jostle
		if Jostle and jostle and self.horizontal then
			if location == "Bottom" then
				Jostle:RegisterTop(self.XPBar)
			else
				Jostle:RegisterBottom(self.XPBar)
			end
		end		
	end
	
	if self:GetSetting("ShowRep") then
		self.RepBar:Show()
		self.RepSpark:Show()
		self.RepSpark2:Show()
		height = height + thickness

		-- register for jostle
		if Jostle and jostle and self.horizontal then
			if location == "Bottom" then
				Jostle:RegisterTop(self.RepBar)
			else
				Jostle:RegisterBottom(self.RepBar)
			end
		end
	end
	
	if self:GetSetting("Shadow") then
		self.Border:Show()
		height = height + 5
	end

	if self.horizontal then
		self.MainFrame:SetHeight(height)
	else
		self.MainFrame:SetWidth(height)		
	end
	
	self.MainFrame:Show() 

	return height
end

function Bar:Hide()
	-- remove from jostle
	if Jostle then
		Jostle:Unregister(self.XPBar)
		Jostle:Unregister(self.RepBar)
	end

	self.XPBar:Hide()
	self.Spark:Hide()
	self.Spark2:Hide()
	self.Border:Hide()
	self.RepBar:Hide()
	self.RepSpark:Hide()
	self.RepSpark2:Hide()
	
	self.MainFrame:Hide() 
end

function Bar:SetThickness(thickness)
	if self.horizontal then
		self.XPBar:SetHeight(thickness)
		self.Spark:SetHeight(thickness * 2 + 12)
		self.Spark2:SetHeight(thickness * 8)

		self.RepBar:SetHeight(thickness)
		self.RepSpark:SetHeight(thickness * 2 + 12)
		self.RepSpark2:SetHeight(thickness * 8)
	else
		self.XPBar:SetWidth(thickness)
		self.Spark:SetWidth(thickness * 2 + 12)
		self.Spark2:SetWidth(thickness * 8)

		self.RepBar:SetWidth(thickness)
		self.RepSpark:SetWidth(thickness * 2 + 12)
		self.RepSpark2:SetWidth(thickness * 8)
	end
end

function Bar:UpdateColor(id)
	local r, g, b, a = self:GetColor(id)

	local tex
	local spark
	
	if id == "XP" then
		tex   = self.XPBarTex
		spark = self.Spark
	elseif id == "Rest" then
		tex = self.RestedXPTex
	elseif id == "None" then
		tex = self.NoXPTex
	elseif id == "Rep" then
		tex   = self.RepBarTex
		spark = self.RepSpark
	elseif id == "NoRep" then
		tex = self.NoRepTex
	end
	
	if tex ~= nil then
		tex:SetVertexColor(r, g, b, a)
	end

	if spark ~= nil then
		spark:SetVertexColor(r, g, b, self:GetSetting("Spark"))
	end
end

function Bar:UpdateSparkIntensity()
	local intensity = self:GetSetting("Spark")
	
	self.Spark:SetAlpha(intensity)
	self.Spark2:SetAlpha(intensity)
	self.RepSpark:SetAlpha(intensity)
	self.RepSpark2:SetAlpha(intensity)
end

function Bar:UpdateStrata()
	local strata = self:GetSetting("Strata")
	
	self.MainFrame:SetFrameStrata(strata)
	self.XPBar:SetFrameStrata(strata)
	self.RepBar:SetFrameStrata(strata)
	self.Border:SetFrameStrata(strata)
end

function Bar:UpdateTexture()
	if not LibSharedMedia then
		return
	end
	
	local texture = self:GetSetting("Texture")
	
	texture = LibSharedMedia:Fetch("statusbar", texture)

	if not texture then
		texture = TEX_BAR
	end

	self.XPBarTex:SetTexture(texture)
	self.NoXPTex:SetTexture(texture)
	self.RestedXPTex:SetTexture(texture)
	self.RepBarTex:SetTexture(texture)
	self.NoRepTex:SetTexture(texture)
end

-- set bar progress in fraction of 1
function Bar:SetBarProgress(bar, fraction)
	if not fraction then
		return
	end
	
	if fraction < 0 then
		fraction = 0
	elseif fraction > 1 and bar ~= "Rested" then
		fraction = 1
	end
	
	self.progress[bar] = fraction
end

-- settings
function Bar:GetSetting(option)
	return self.settings[option]
end

function Bar:SetSetting(option, value, skipReanchor)
	local current = self:GetSetting(option)

	if current == value then
		return
	end
	
	self.settings[option] = value
	
	if not skipReanchor then
		self:Reanchor()
	end
end

function Bar:GetColor(id)
	local color = self.colors[id] or {r = 0, g = 0, b = 0, a = 0}

	return color.r, color.g, color.b, color.a
end

function Bar:SetColor(id, r, g, b, a)
	if not self.colors[id] then
		return
	end

	self.colors[id].r = r
	self.colors[id].g = g
	self.colors[id].b = b
	self.colors[id].a = a
	
	self:UpdateColor(id)
end
