local _G = _G

local ADDON = ...

local BrokerXPBar = _G.BrokerXPBar

-- tooltip library
local QT		= LibStub:GetLibrary("LibQTip-1.0")

-- get translations
local L         = LibStub:GetLibrary("AceLocale-3.0"):GetLocale( ADDON )

-- coloring tools
local Crayon	= LibStub:GetLibrary("LibCrayon-3.0")

-- local functions
local floor     = math.floor

local UnitXP          = _G.UnitXP
local UnitXPMax       = _G.UnitXPMax
local GetFactionInfo  = _G.GetFactionInfo
local GetXPExhaustion = _G.GetXPExhaustion

-- colors
local function Red( text )      return "|cffff0000" .. text .. "|r" end
local function White( text )    return "|cffffffff" .. text .. "|r" end
local function Yellow( text )   return "|cffffff00" .. text .. "|r" end
local function Brownish( text ) return "|cffeda55f" .. text .. "|r" end

local tooltip

function BrokerXPBar:CreateTooltip( obj )
	tooltip = QT:Acquire( ADDON.."TT", 2 )
	tooltip:Hide()
	tooltip:Clear()
	tooltip:SetScale(1)
		
	self:DrawTooltip()

	tooltip:SmartAnchorTo( obj )
	tooltip:Show()
end

function BrokerXPBar:RemoveTooltip()
	if tooltip then
		tooltip:Hide()
		QT:Release(tooltip)
		tooltip = nil
	end
end

function BrokerXPBar:DrawTooltip()
	local colcount = tooltip:GetColumnCount()	
	
	local totalXP = UnitXPMax("player")
	local currentXP = UnitXP("player")
	local toLevelXP = totalXP - currentXP
	local toLevelXPPercent = floor((currentXP / totalXP) * 100)

	-- Add our header
	local lineNum = tooltip:AddHeader( " " )
	tooltip:SetCell( lineNum, 1, White( self.FULLNAME ), "CENTER", colcount )
	tooltip:AddLine( " " )

	if self:GetSetting("ShowXP") then
		local exhaustion = GetXPExhaustion()
		if exhaustion then
			local xpEx = exhaustion
			local xpExPercent
			if xpEx - toLevelXP > 0 then
				xpExPercent = floor(((xpEx - toLevelXP) / totalXP) * 100)
			else
				xpExPercent = floor((xpEx / totalXP) * 100)
			end
			if Crayon then
				-- Scale: 1 - 100
				-- ExXP:  1 - toLevelXP
				xpEx = "|cff"..Crayon:GetThresholdHexColor(xpEx, 1, toLevelXP * 0.25, toLevelXP * 0.5, toLevelXP * 0.75, toLevelXP) .. self:FormatNumberTip(xpEx) .. "|r"
				xpExPercent = "|cff"..Crayon:GetThresholdHexColor(xpExPercent, 1, 25, 50, 75, 100) .. self:FormatNumberTip(xpExPercent) .. "|r"
			else
				xpEx = self:FormatNumberTip(xpEx, true)
				xpExPercent = self:FormatNumberTip(xpExPercent, true)
			end
			if exhaustion - toLevelXP > 0 then
				xpExPercent = "100% + "..xpExPercent
			end
			lineNum = tooltip:AddLine( " " )
			tooltip:SetCell( lineNum, 1, L["Rested XP"], "LEFT" )
			tooltip:SetCell( lineNum, 2, string.format("%s (%s%%)", xpEx, xpExPercent), "LEFT" )
		end
		if Crayon then
			currentXP = "|cff"..Crayon:GetThresholdHexColor(currentXP, 1, totalXP * 0.25, totalXP * 0.5, totalXP * 0.75, totalXP) .. self:FormatNumberTip(currentXP) .. "|r"
			toLevelXP = "|cff"..Crayon:GetThresholdHexColor(toLevelXP, totalXP, totalXP * 0.75, totalXP * 0.5, totalXP * 0.25, 1) .. self:FormatNumberTip(toLevelXP) .. "|r"
			toLevelXPPercent = "|cff"..Crayon:GetThresholdHexColor(toLevelXPPercent, 1, 25, 50, 75, 100) .. toLevelXPPercent .. "|r"
		else
			currentXP = self:FormatNumberTip(currentXP, true)
			toLevelXP = self:FormatNumberTip(toLevelXP, true)
		end
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Current XP"], "LEFT" )
		tooltip:SetCell( lineNum, 2, string.format("%s/%s (%s%%)", currentXP, self:FormatNumberTip(totalXP), toLevelXPPercent), "LEFT" )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["To Level"], "LEFT" )
		tooltip:SetCell( lineNum, 2, toLevelXP, "LEFT" )
	end
	
	if self:GetSetting("ShowRep") and self.faction ~= 0 then
		local name, desc, standing, minRep, maxRep, currentRep = GetFactionInfo(self.faction)
		local r, g, b, a = self:GetBlizzardReputationColor(standing)
		
		local fullLevelRep = maxRep - minRep
		local toLevelRep   = maxRep - currentRep
		local atLevelRep   = fullLevelRep - toLevelRep

		if Crayon then
			toLevelRep = "|cff"..Crayon:GetThresholdHexColor(atLevelRep, 1, fullLevelRep * 0.25, fullLevelRep * 0.5, fullLevelRep * 0.75, fullLevelRep) .. self:FormatNumberTip(toLevelRep) .. "|r"
		else
			toLevelRep = self:FormatNumberTip(toLevelRep)
		end

		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Faction"], "LEFT" )
		tooltip:SetCell( lineNum, 2, name, "LEFT" )

		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Rep to next standing"], "LEFT" )
		tooltip:SetCell( lineNum, 2, toLevelRep, "LEFT" )

		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Current rep"], "LEFT" )
		tooltip:SetCell( lineNum, 2, "|cff"..string.format("%02x%02x%02x", r*255, g*255, b*255)..getglobal("FACTION_STANDING_LABEL"..standing).."|r", "LEFT" )
	end

	if self:GetSetting("ShowXP") and self.playerLvl < self.MAX_LEVEL then
		self.History:Process()
		
		local kph  = self.History:GetKillsPerHour()
		local xpph = self.History:GetXPPerHour()
		
		if kph then
			kph = string.format("%.0f", kph)
		else
			kph = "~"
		end

		if xpph then
			xpph = string.format("%.0f", xpph)
		else
			xpph = "~"
		end

		tooltip:AddLine( " " )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Session XP"], "LEFT" )
		tooltip:SetCell( lineNum, 2, self.History.totalXP, "LEFT" )
		
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Session Kills"], "LEFT" )
		tooltip:SetCell( lineNum, 2, self.History.totalKills, "LEFT" )
		
		tooltip:AddLine( " " )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["XP per hour"], "LEFT" )
		tooltip:SetCell( lineNum, 2, xpph, "LEFT" )

		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Kills per hour"], "LEFT" )
		tooltip:SetCell( lineNum, 2, kph, "LEFT" )
		
		tooltip:AddLine( " " )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Time to level"], "LEFT" )
		tooltip:SetCell( lineNum, 2, self.History:GetTimeToLevel(), "LEFT" )
		
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, L["Kills to level"], "LEFT" )
		tooltip:SetCell( lineNum, 2, Red( self.History:GetKillsToLevel() ), "LEFT" )		
	end
	
	if not self:GetSetting("HideHint") then
		tooltip:AddLine( " " )
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1,  Brownish( L["Click"] .. ":" ) .. Yellow( " " .. L["Send current XP to an open editbox."] ), nil, "LEFT", colcount)
		if self.faction ~= 0 then
			lineNum = tooltip:AddLine( " " )
			tooltip:SetCell( lineNum, 1, Brownish( L["Shift-Click"] .. ":" ) .. Yellow( " " .. L["Send current reputation to an open editbox."] ), nil, "LEFT", colcount )
		end
		lineNum = tooltip:AddLine( " " )
		tooltip:SetCell( lineNum, 1, Brownish( L["Right-Click"] .. ":" ) .. Yellow( " " .. L["Open option menu."] ), nil, "LEFT", colcount )
	end
	
end

